/*
Copyright (C) THL A29 Limited, a Tencent company. All rights reserved.

SPDX-License-Identifier: Apache-2.0
*/

package dpos

import (
	"encoding/binary"
	"fmt"
	"math/rand"
	"sort"
	"strings"

	"chainmaker.org/chainmaker/pb-go/v2/consensus"

	"chainmaker.org/chainmaker/vm-native/v2/dposmgr"

	configPb "chainmaker.org/chainmaker/pb-go/v2/config"
	pbdpos "chainmaker.org/chainmaker/pb-go/v2/consensus/dpos"
	"chainmaker.org/chainmaker/pb-go/v2/syscontract"
	"chainmaker.org/chainmaker/protocol/v2"
	"chainmaker.org/chainmaker/utils/v2"
	"github.com/gogo/protobuf/proto"
	goproto "github.com/gogo/protobuf/proto"
)

// ValidatorsElection select validators from Candidates
func ValidatorsElection(
	infos []*pbdpos.CandidateInfo, n int, seed []byte, outSort bool) ([]*pbdpos.CandidateInfo, error) {
	if n == 0 {
		return nil, fmt.Errorf("can not select zero validators")
	}
	m := len(infos)
	if n > m {
		return nil, fmt.Errorf("the number of candidate is not enough, candidate[%v] validator[%v]", m, n)
	}
	if n == m {
		validators := make([]*pbdpos.CandidateInfo, 0)
		validators = append(validators, infos...)
		return validators, nil
	}
	// when m > n
	// 首先对结果进行排序
	sort.Sort(CandidateInfos(infos))
	// n、m 都分为 两部分，
	var (
		m0, _      = distributionM(m, n)
		n0, n1     = distributionN(m, n)
		validators = make([]*pbdpos.CandidateInfo, 0)
	)
	// 首选从m0个对象中选择n0个结果
	// |m0|after(m0) ~ m1|
	// |n0|n1|
	seedInt := binary.LittleEndian.Uint64(seed)
	// 设置种子
	rand.Seed(int64(seedInt))
	selectM0IdxMap := sliceToMap(rand.Perm(m0)[:n0])
	for k := range selectM0IdxMap {
		validators = append(validators, infos[k])
	}
	// 构建新的数组
	newSelectArray := make([]*pbdpos.CandidateInfo, 0)
	for i := 0; i < len(infos); i++ {
		if _, ok := selectM0IdxMap[i]; !ok {
			newSelectArray = append(newSelectArray, infos[i])
		}
	}

	// 从新的数组中选择n1个结果
	// 理论上不需要再排序，因为最开始已经排序过
	rand.Seed(int64(seedInt)) // 设置种子
	selectM1IdxArray := rand.Perm(len(newSelectArray))[:n1]
	for _, v := range selectM1IdxArray {
		validators = append(validators, newSelectArray[v])
	}
	if outSort {
		// 输出需要进行排序
		sort.Sort(CandidateInfos(validators))
	}
	return validators, nil
}

// n、m 都分为 两部分
func distributionM(m, n int) (int, int) {
	return n, m - n
}

// n、m 都分为 两部分，
func distributionN(m, n int) (int, int) {
	n0 := n / 2
	return n0, n - n0
}

// type conversion
func sliceToMap(array []int) map[int]struct{} {
	values := make(map[int]struct{})
	for _, v := range array {
		values[v] = struct{}{}
	}
	return values
}

// CandidateInfos array for sort
type CandidateInfos []*pbdpos.CandidateInfo

// Len return length
func (s CandidateInfos) Len() int {
	return len(s)
}

// Swap swap two elements
func (s CandidateInfos) Swap(i, j int) {
	s[i], s[j] = s[j], s[i]
}

// Less compare two elements
func (s CandidateInfos) Less(i, j int) bool {
	// 优先按照weight排序，相同的情况下按照PeerId从小到大排序（字符串）
	wi, wj := utils.NewBigInteger(s[i].Weight), utils.NewBigInteger(s[j].Weight)
	val := wi.Cmp(wj)
	if val == 0 {
		return strings.Compare(s[i].PeerId, s[j].PeerId) < 0
	}
	return val > 0
}

// getLatestEpochInfo get information on the last epoch
func getLatestEpochInfo(store protocol.BlockchainStore) (*syscontract.Epoch, error) {
	//returns the state value for given contract name and key, or returns nil if none exists.
	val, err := store.ReadObject(syscontract.SystemContract_DPOS_STAKE.String(), []byte(dposmgr.KeyCurrentEpoch))

	if err != nil {
		return nil, fmt.Errorf("read contract: %s key: %s, error: %s",
			syscontract.SystemContract_DPOS_STAKE.String(), dposmgr.KeyCurrentEpoch, err)
	}
	epoch := syscontract.Epoch{}
	if err = proto.Unmarshal(val, &epoch); err != nil {
		return nil, fmt.Errorf("unmarshal epoch failed, reason: %s", err)
	}
	return &epoch, nil
}

// getNodeIDsFromValidators query the corresponding node id of the validator from the ledger,
// and return the list of node ids
func getNodeIDsFromValidators(store protocol.BlockchainStore, validators []string) ([]string, error) {
	if len(validators) == 0 {
		return nil, fmt.Errorf("validators is null")
	}
	nodeIDs := make([]string, 0, len(validators))
	for _, validator := range validators {
		//returns the state value for given contract name and key, or returns nil if none exists.
		nodeID, err := store.ReadObject(syscontract.SystemContract_DPOS_STAKE.String(), dposmgr.ToNodeIDKey(validator))
		if err != nil || len(nodeID) == 0 {
			return nil, fmt.Errorf("read nodeID of the validator[%s] failed, reason: %s", validator, err)
		}
		nodeIDs = append(nodeIDs, string(nodeID))
	}
	return nodeIDs, nil
}

// getChainConfig query chain configuration from ledger
func getChainConfig(store protocol.BlockchainStore) (*configPb.ChainConfig, error) {
	var chainConfig configPb.ChainConfig
	//returns the state value for given contract name and key, or returns nil if none exists.
	bytes, err := store.ReadObject(
		syscontract.SystemContract_CHAIN_CONFIG.String(),
		[]byte(syscontract.SystemContract_CHAIN_CONFIG.String()),
	)
	if err != nil || len(bytes) == 0 {
		return nil, fmt.Errorf("read chainConfig failed, reason: %s", err)
	}
	err = goproto.Unmarshal(bytes, &chainConfig)
	if err != nil {
		err = fmt.Errorf(
			"unmarshal chainConfig failed, contractName %s err: %+v",
			syscontract.SystemContract_CHAIN_CONFIG.String(),
			err,
		)
		return nil, err
	}
	return &chainConfig, nil
}

// GetValidatorList get validator list, used by chainmaker-go
// Only used for dpos consensus,
// returns an error when configuring other consensus algorithms
// nolint: unused
func GetValidatorList(chainConfig *configPb.ChainConfig, store protocol.BlockchainStore) (
	validators []string, err error) {
	if chainConfig.Consensus.Type == consensus.ConsensusType_DPOS {
		return getValidatorListFromLedger(store)
	}
	return nil, fmt.Errorf("unknown consensus type: %s", chainConfig.Consensus.Type)
}

// getValidatorListFromLedger get validator list from ledger
func getValidatorListFromLedger(store protocol.BlockchainStore) (validators []string, err error) {
	epoch, err := getLatestEpochInfo(store)
	if err != nil {
		return nil, err
	}
	// query the corresponding node id of the validator from the ledger,
	// and return the list of node ids
	nodeIDs, err := getNodeIDsFromValidators(store, epoch.ProposerVector)
	if err != nil {
		return nil, err
	}
	return nodeIDs, nil
}
